package mngr

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
	"gitlab.smartblocklab.com/smartcontracts/common/model/offerCC"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
)

func (this OfferSessionManagerImpl) SaveOfferPair(outOfferRequest *offerCC.OfferRecord, inOfferRequest *offerCC.OfferRecord) (*offerCC.OfferRecords, error) {
	if outOfferRequest == nil {
		return nil, errors.New("out OfferRecord is nil")
	}
	if inOfferRequest == nil {
		return nil, errors.New("in OfferRecord is nil")
	}

	outAccount, err := this.AccountCC.CheckAndLoadAccount(outOfferRequest.UserAccountInfo)
	if err != nil {
		return nil, err
	}
	err = outAccount.Validate()
	if err != nil {
		return nil, err
	}

	inAccount, err := this.AccountCC.CheckAndLoadAccount(inOfferRequest.UserAccountInfo)
	if err != nil {
		return nil, err
	}
	err = inAccount.Validate()
	if err != nil {
		return nil, err
	}

	if outAccount.UserID != outOfferRequest.UserAccountInfo.UserID {
		return nil, errors.Errorf("User can't create offer for self. OutUserInfo.UserID %s inAccount.UserID %s",
			outOfferRequest.UserAccountInfo.UserID, inAccount.UserID)
	}

	if outAccount.UserID == inAccount.UserID {
		return nil, errors.Errorf("User can't create offer for self. OutUserInfo.UserID %s inAccount.UserID %s",
			outAccount.UserID, inAccount.UserID)
	}

	if outAccount.Status != accountCC.AccountStatus_ACTIVE {
		return nil, errors.Errorf("out account status invalid. Expected Active, actual status %s", outAccount.Status)
	}

	if inAccount.Status != accountCC.AccountStatus_ACTIVE {
		return nil, errors.Errorf("in account status invalid. Expected Active, actual status %s", inAccount.Status)
	}

	// update loaded account
	outOfferRequest.UserAccountInfo = outAccount.ToUserAccountInfo()
	inOfferRequest.UserAccountInfo = inAccount.ToUserAccountInfo()

	inOfferRequest.Key, err = this.CreateCompositeKeyByValidObject(inOfferRequest)
	if err != nil {
		return nil, err
	}

	outOfferRequest.Key, err = this.CreateCompositeKeyByValidObject(outOfferRequest)
	if err != nil {
		return nil, err
	}

	outOfferRequest.DependencyKey = inOfferRequest.Key
	inOfferRequest.DependencyKey = outOfferRequest.Key

	err = outOfferRequest.Validate()
	if err != nil {
		return nil, err
	}

	err = inOfferRequest.Validate()
	if err != nil {
		return nil, err
	}

	if outOfferRequest.OperationID != inOfferRequest.OperationID {
		return nil, logging.Get().Errorf("OperationID out %s in %s not equals",
			outOfferRequest.OperationID, inOfferRequest.OperationID)
	}
	if outOfferRequest.OfferRecordStatus != inOfferRequest.OfferRecordStatus {
		return nil, logging.Get().Errorf("OfferRequestStatus out %s in %s not equals",
			outOfferRequest.OfferRecordStatus, inOfferRequest.OfferRecordStatus)
	}

	err = this.checkLockAmountByOperationIDAndUserID(outOfferRequest)
	if err != nil {
		return nil, err
	}

	if len(outOfferRequest.ReplayedKey) != 0 && outOfferRequest.OfferRecordStatus != offerCC.OfferRecordStatusReplay {
		bytes, err := this.GetState(outOfferRequest.ReplayedKey)
		if err != nil {
			return nil, err
		}

		replOfferRecord, err := offerCC.Unmarshal(bytes)
		err = this.checkLockAmountByOperationIDAndUserID(replOfferRecord)
		if err != nil {
			return nil, err
		}

		bytes, err = this.GetState(replOfferRecord.DependencyKey)
		if err != nil {
			return nil, err
		}

		replDepOfferRecord, err := offerCC.Unmarshal(bytes)
		err = this.checkLockAmountByOperationIDAndUserID(replDepOfferRecord)
		if err != nil {
			return nil, err
		}
	}

	chaincode.GlobalLogger.Debugf("Adding new offer offerRequest from user %s, operation id %s",
		outOfferRequest.UserAccountInfo.UserID, outOfferRequest.OperationID)
	chaincode.GlobalLogger.Debugf("Adding new offer offerRequest to user %s, operation id %s",
		inOfferRequest.UserAccountInfo.UserID, inOfferRequest.OperationID)

	//chaincode.GlobalLogger.Debugf("adding new offer into ledger, operation id %s, out user id %s, from currency id %s",
	//	offer.OperationID, offer.OutUserInfo.UserID, offer.OutUserInfo.Amount.CurrencyType)

	// Adding new offer into ledger state

	result := &offerCC.OfferRecords{
		OfferRecords: make([]*offerCC.OfferRecord, 0),
	}
	outValidObject, err := this.AddObject(outOfferRequest)
	if err != nil {
		return nil, err
	}
	outAddOfferRecord, err := offerCC.Cast(outValidObject)
	result.OfferRecords = append(result.OfferRecords, outAddOfferRecord)

	inValidObject, err := this.AddObject(inOfferRequest)
	if err != nil {
		return nil, err
	}
	inAddOfferRecord, err := offerCC.Cast(inValidObject)
	result.OfferRecords = append(result.OfferRecords, inAddOfferRecord)

	return result, nil
}
