package chaincode

import (
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
	"gitlab.smartblocklab.com/smartcontracts/common/security/actions"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
)

type InternalUser struct {
	UserID string          `json:"user_id"`
	Roles  []actions.Roles `json:"roles"`
}

func CheckAccessParameterByArgsAndFunctionName(channelName string, chaincodeName string, args []string, functionName string, manager IManager) (*requests.Parameter, *requests.Parameter, error) {
	preferences, err := actions.FindSecurityPreferenceByFunctionName(chaincodeName, functionName)
	if err != nil {
		logging.Get().Error(err.Error())
		return nil, nil, err
	}
	logging.Get().Infof("preferences.Function: %s", preferences.FunctionName)
	logging.Get().Infof("preferences.Chaincode: %s", preferences.ChaincodeName)
	logging.Get().Infof("preferences.ParameterType: %s", preferences.ParameterType)
	return CheckAccessParameter(channelName, manager, args, preferences)
}

/**
@return UserParams, SystemParams, error
*/
func CheckAccessParameter(channelName string, manager IManager, args []string, preferences actions.SecurityPreference) (*requests.Parameter, *requests.Parameter, error) {
	logging.Get().Infof("CheckAccess: chaincode=%s, method=%s", preferences.ChaincodeName, preferences.FunctionName)

	if len(args) != 1 {
		return nil, nil, logging.Get().Errorf("CheckAccess: Incorrect number of arguments, expecting %d, while got %d", 1, len(args))
	}

	logging.Get().Infof("FullSignedRequest: %s", args[0])
	var signedMessage requests.FullSignedRequest
	if err := json.Unmarshal([]byte(args[0]), &signedMessage); err != nil {
		return nil, nil, err
	}

	var userParam *requests.Parameter
	var systemParam *requests.Parameter
	var err error

	if preferences.ParameterType == actions.UserParameter || preferences.ParameterType == actions.ComplexParameter {
		logging.Get().Infof("CheckAccess: user method: %s.%s", preferences.ChaincodeName, preferences.FunctionName)

		userParam, err = verifyParameter(channelName, manager, preferences, signedMessage.Parameters.UserParams, signedMessage.UserSignature);
		if err != nil {
			return nil, nil, logging.Get().Errorf("CheckAccess: parameter verification failed: %s", err.Error())
		}

		if !checkRole(manager, userParam, preferences) {
			return nil, nil, logging.Get().Errorf("CheckAccess: Role checking failed")
		}

		if preferences.ParameterType == actions.UserParameter {
			logging.Get().Infof("CheckAccess: [UserParameter] Access granted for user '%s'", userParam.UserID)
			return userParam, nil, nil
		}

		// verify system params
		if systemParam, err = verifyParameter(channelName, manager, preferences, signedMessage.Parameters.SystemParams, signedMessage.SystemSignature); err != nil {
			return nil, nil, logging.Get().Errorf("CheckAccess: verification failed: %s", err.Error())
		}

		if !checkRole(manager, systemParam, preferences) {
			return nil, nil, logging.Get().Error("CheckAccess: access denied: system user has no role")
		}

		logging.Get().Infof("CheckAccess: chaincode=%s, method=%s. [UserParameter, ServiceParameter] Access granted for user '%s' and system user '%s'",
			preferences.ChaincodeName, preferences.FunctionName, userParam.UserID, systemParam.UserID)

		return userParam, systemParam, nil

	} else if preferences.ParameterType == actions.ServiceParameter {
		logging.Get().Infof("CheckAccess: service method: %s.%s", preferences.ChaincodeName, preferences.FunctionName)
		// verify system params
		if systemParam, err = verifyParameter(channelName, manager, preferences, signedMessage.Parameters.SystemParams, signedMessage.SystemSignature); err != nil {
			return nil, nil, logging.Get().Errorf("CheckAccess: verification failed: %s", err.Error())
		}

		if !checkRole(manager, systemParam, preferences) {
			return nil, nil, logging.Get().Error("CheckAccess: access denied: system user has no role")
		}

		logging.Get().Infof("CheckAccess: chaincode=%s, method=%s. [ServiceParameter] Access granted for system user '%s'",
			preferences.ChaincodeName, preferences.FunctionName, systemParam.UserID)

		return nil, systemParam, nil

	} else if preferences.ParameterType == actions.ServiceParameter|actions.UnsecuredParameter {
		logging.Get().Infof("CheckAccess: unsecured service method: %s.%s", preferences.ChaincodeName, preferences.FunctionName)

		if err = json.Unmarshal(signedMessage.Parameters.SystemParams, &systemParam); err != nil {
			return nil, nil, logging.Get().Errorf("CheckAccess: Unmarshaling error: %s", err.Error())
		}
		logging.Get().Infof("CheckAccess: chaincode=%s, method=%s. [Unsecured service Parameter] Access granted for system user '%s'",
			preferences.ChaincodeName, preferences.FunctionName, systemParam.UserID)

		return nil, systemParam, nil

	} else if preferences.ParameterType == actions.UserParameter|actions.UnsecuredParameter {
		logging.Get().Infof("CheckAccess: unsecured user method: %s.%s", preferences.ChaincodeName, preferences.FunctionName)

		if err = json.Unmarshal(signedMessage.Parameters.UserParams, &userParam); err != nil {
			return nil, nil, logging.Get().Errorf("CheckAccess: Unmarshaling error: %s", err.Error())
		}
		logging.Get().Infof("CheckAccess: chaincode=%s, method=%s. [Unsecured user Parameter] Access granted for system user '%s'",
			preferences.ChaincodeName, preferences.FunctionName, systemParam.UserID)

		return userParam, nil, nil

	} else {
		return nil, nil, logging.Get().Errorf("CheckAccess: Unknown ParameterType: %d ", preferences.ParameterType)
	}
}

func CheckAccess(channelName string, manager IManager, args []string, chaincodeName string, functionName string) (json.RawMessage, json.RawMessage, error) {
	preferences, err := actions.FindSecurityPreferenceByFunctionName(chaincodeName, functionName)
	if err != nil {
		return nil, nil, err
	}

	userParameter, systemParameter, err := CheckAccessParameter(channelName, manager, args, preferences)

	var systemParam json.RawMessage = nil
	if systemParameter != nil {
		systemParam = systemParameter.Param
	}

	var userParam json.RawMessage = nil
	if userParameter != nil {
		userParam = userParameter.Param
	}
	return userParam, systemParam, err
}

func verifyParameter(channelName string, manager IManager, preferences actions.SecurityPreference, rawParameter json.RawMessage, sig requests.UserSignature) (*requests.Parameter, error) {
	logging.Get().Infof("VerifyParameter: ChaincodeName=%s, method=%s, parameter=%s", preferences.ChaincodeName,
		preferences.FunctionName, string(rawParameter))

	var parameter requests.Parameter
	if err := json.Unmarshal(rawParameter, &parameter); err != nil {
		return nil, logging.Get().Errorf("VerifyParameter: rawParameter: %s. Unmarshaling error: %s", string(rawParameter), err.Error())
	}

	if preferences.GetActionID(channelName) != parameter.ActionId {
		return nil, logging.Get().Errorf("VerifyParameter: Action Ids are not equal – called id: %d, caller id: %d",
			preferences.GetActionID(channelName), parameter.ActionId)
	}

	hasher := sha256.New()
	hasher.Write(rawParameter)
	digest := hasher.Sum(nil)
	strDigest := hex.EncodeToString(digest)

	arg := requests.VerifyUserSignatureArg{
		UserID:        parameter.UserID,
		Digest:        strDigest,
		UserSignature: sig,
	}

	logging.Get().Infof("VerifyParameter: Invoke verifyUserSignature method")
	resp := manager.InvokeChaincode_VerifyUserSignature(arg)

	if resp.Status != 200 {
		return nil, logging.Get().Errorf("VerifyParameter: Invoke failed: %s ", resp.Message)
	} else {
		return &parameter, nil
	}
}

func checkRole(manager IManager, parameter *requests.Parameter, preferences actions.SecurityPreference) bool {
	logging.Get().Infof("checkRole: tools.InvokeChaincode_GetUser %s", parameter.UserID)

	resp := manager.InvokeChaincode_GetUser(parameter.UserID)
	logging.Get().Infof("checkRole: response status: %d", resp.Status)
	logging.Get().Infof("checkRole: payload: %s", string(resp.Payload))
	logging.Get().Infof("checkRole: message: %s", string(resp.Message))

	if resp.Status != 200 {
		logging.Get().Warnf("checkRole: Invoke chaincode userCC.getUser by userID %s failed, with response status: %s, message: %s, payload: %s",
			parameter.UserID, resp.Status, resp.Message, resp.String())
		return false

	} else {
		var user InternalUser
		if err := json.Unmarshal(resp.Payload, &user); err != nil {
			logging.Get().Warnf("checkRole: Checking roles failed, can't unmarshal user: %s ", resp.Message)
			return false
		}

		logging.Get().Infof("checkRole: ChaincodeName=%s, method=%s, user=%s", preferences.ChaincodeName, preferences.FunctionName, string(resp.Payload))

		if parameter.UserID == user.UserID {
			logging.Get().Info("checkRole: Found same Parameter UserID and Fetched UserID")

			if len(user.Roles) == 0 && len(preferences.Roles) == 0 {
				return true
			}

			for _, userRole := range user.Roles {
				for _, methodRole := range preferences.Roles {
					if userRole == methodRole {
						logging.Get().Info("checkRole: Role is matched")
						return true
					}
				}
			}
		}

		logging.Get().Error("checkRole: User has no any suitable role")
		return false
	}
}
