package chaincode

import (
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"encoding/json"
	"github.com/pkg/errors"
	"reflect"
	"gitlab.smartblocklab.com/smartcontracts/common/security/actions"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"strings"
)

// Пример наследования данной структуры
//type AccountManagerImpl struct {
//	*manager.ManagerImpl
//}

func GetManager(adapter StubInterfaceAdapter, timeGetter TimeGetter) IManager {
	GlobalLogger.Infof("GetManager")
	manager := &ManagerImpl{adapter, timeGetter}
	return manager
}

type ManagerImpl struct {
	Adapter    StubInterfaceAdapter
	TimeGetter TimeGetter
}

func (this ManagerImpl) GetTime() int64 {
	GlobalLogger.Infof("managerImpl.GetTime")
	tt, err := this.TimeGetter()
	if err != nil {
		GlobalLogger.Error(err.Error())
		return 0
	} else {
		return tt
	}
}

func (this ManagerImpl) SetAdapter(new StubInterfaceAdapter) {
	GlobalLogger.Infof("managerImpl.SetAdapter")
	this.Adapter = new
}

func (this ManagerImpl) SetTimeGetter(new TimeGetter) {
	GlobalLogger.Infof("managerImpl.SetTimeGetter")
	this.TimeGetter = new
}

func (this ManagerImpl) GetAdapter() StubInterfaceAdapter {
	GlobalLogger.Infof("managerImpl.GetAdapter")
	return this.Adapter
}

func (this ManagerImpl) GetTimeGetter() TimeGetter {
	GlobalLogger.Infof("managerImpl.GetTimeGetter")
	return this.TimeGetter
}

func (this ManagerImpl) GetState(key string) ([]byte, error) {
	GlobalLogger.Infof("managerImpl.GetState")
	return this.Adapter.GetState(key)
}

func (this ManagerImpl) PutState(key string, value []byte) error {
	GlobalLogger.Infof("managerImpl.PutState by key %s", key)
	return this.Adapter.PutState(key, value)
}

func (this ManagerImpl) CreateCompositeKeyByValidObject(validObject ValidObject) (string, error) {
	objectType := validObject.GetObjectType()
	key := validObject.GetKey(this)
	GlobalLogger.Infof("managerImpl.CreateCompositeKeyByValidObject by objectType %s and attributes %v", objectType, key)
	return this.Adapter.CreateCompositeKey(objectType, key)
}

func (this ManagerImpl) CreateCompositeKey(objectType string, attributes []string) (string, error) {
	GlobalLogger.Infof("managerImpl.CreateCompositeKey by objectType %s and attributes %v", objectType, attributes)
	return this.Adapter.CreateCompositeKey(objectType, attributes)
}

func (this ManagerImpl) GetStateByPartialCompositeKey(objectType string, keys []string) (shim.StateQueryIteratorInterface, error) {
	GlobalLogger.Infof("managerImpl.GetStateByPartialCompositeKey by objectType %s and attributes %v", objectType, keys)
	return this.Adapter.GetStateByPartialCompositeKey(objectType, keys)
}

func (this ManagerImpl) GetObjectListByKeys(expectedValidObjectType ValidObject, keys []string) ([]ValidObject, error) {
	objectType := expectedValidObjectType.GetObjectType()
	GlobalLogger.Infof("managerImpl.GetObjectListByPartialCompositeKey by objectType %s and attributes %v", objectType, keys)
	it, err := this.GetStateByPartialCompositeKey(objectType, keys)
	if err != nil {
		return nil, err
	}

	defer it.Close()

	validObjectArray := make([]ValidObject, 0)

	// Iterate through result set and for each marble found, transfer to newOwner
	var i int
	for i = 0; it.HasNext(); i++ {
		// Note that we don't get the value (2nd return variable), we'll just get the marble name from the composite key
		responseRange, err := it.Next()
		if err != nil {
			return nil, err
		}

		validObject := expectedValidObjectType.New()

		err = json.Unmarshal(responseRange.Value, validObject)
		if err != nil {
			return nil, GlobalLogger.Error(err.Error())
		}

		validObjectArray = append(validObjectArray, validObject)
	}
	return validObjectArray, nil
}

// поиск списка объектов по ключам и фильтр по массиву строк
func (this ManagerImpl) GetObjectListByKeysWithFilter(expectedValidObjectType ValidObject, keys []string, filterArray...string) ([]ValidObject, error) {
	objectType := expectedValidObjectType.GetObjectType()
	GlobalLogger.Infof("managerImpl.GetObjectListByPartialCompositeKey by objectType %s and attributes %v", objectType, keys)
	it, err := this.GetStateByPartialCompositeKey(objectType, keys)
	if err != nil {
		return nil, err
	}

	defer it.Close()

	validObjectArray := make([]ValidObject, 0)

	// Iterate through result set and for each marble found, transfer to newOwner
	var i int
	for i = 0; it.HasNext(); i++ {
		// Note that we don't get the value (2nd return variable), we'll just get the marble name from the composite key
		responseRange, err := it.Next()
		if err != nil {
			return nil, err
		}

		validObject := expectedValidObjectType.New()

		err = json.Unmarshal(responseRange.Value, validObject)
		if err != nil {
			return nil, GlobalLogger.Error(err.Error())
		}

		if isMatchedFilter(string(responseRange.Value), filterArray) {
			validObjectArray = append(validObjectArray, validObject)
		}
	}
	return validObjectArray, nil
}

func isMatchedFilter(sourceString string, filterArray []string) bool {
	matchFound := 0
	for _, f := range filterArray {
		GlobalLogger.Info("check filter")
		GlobalLogger.Info(f)
		if !strings.Contains(sourceString, f) {
			GlobalLogger.Infof("not contains %s in %s\n/n", f, sourceString)
			break
		}
		matchFound++
	}
	return matchFound == len(filterArray)
}

func (this ManagerImpl) SplitCompositeKey(compositeKey string) (string, []string, error) {
	GlobalLogger.Infof("managerImpl.SplitCompositeKey by key %s", compositeKey)
	return this.Adapter.SplitCompositeKey(compositeKey)
}

func (this ManagerImpl) DelState(key string) error {
	GlobalLogger.Infof("managerImpl.DelState %s", key)
	return this.Adapter.DelState(key)
}

func (this ManagerImpl) InvokeChaincode(chaincodeName string, args [][]byte, channel string) pb.Response {
	GlobalLogger.Infof("managerImpl.InvokeChaincode")
	return this.Adapter.InvokeChaincode(chaincodeName, args, channel)
}

func (this ManagerImpl) GetTxID() string {
	GlobalLogger.Infof("managerImpl.GetTxID")
	return this.Adapter.GetTxID()
}

func (this ManagerImpl) Save(obj ValidObject) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.Save")
	if obj == nil {
		return nil, errors.New("can't save nil object")
	}
	if reflect.ValueOf(obj).Kind() != reflect.Ptr {
		return nil, errors.New("check newObject is not reference")
	}

	err := obj.Validate()
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	objByteArray, err := json.Marshal(obj)
	if err != nil {
		return nil, GlobalLogger.Errorf("failed to Marshal. error %s", err)
	}

	// Convert keys to  compound key
	compositeKey, err := this.CreateCompositeKeyByValidObject(obj)
	if err != nil {
		return nil, GlobalLogger.Errorf("Failed to create composite key. Error: %s", err)
	}

	// Add Object JSON to state
	err = this.PutState(compositeKey, objByteArray)
	if err != nil {
		return nil, GlobalLogger.Errorf("Failed to PutState. Error: %s", err)
	}

	return obj, nil
}

func (this ManagerImpl) IsExists(obj ValidObject) (bool, error) {
	GlobalLogger.Infof("managerImpl.IsExists")
	compositeKey, err := this.CreateCompositeKeyByValidObject(obj)
	if err != nil {
		return false, GlobalLogger.Errorf("Failed to create composite key. Error: %s", err)
	}

	bytes, err := this.GetState(compositeKey)
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	if bytes == nil || len(bytes) == 0 {
		return false, nil
	}

	err = json.Unmarshal(bytes, &obj)
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	return true, nil
}

func (this ManagerImpl) GetIfExists(newObject ValidObject) (bool, error) {
	GlobalLogger.Infof("managerImpl.GetIfExists")
	bytes, err := this.GetObject(newObject.New(), newObject.GetKey(this))
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	return bytes != nil, nil
}

func (this ManagerImpl) AddObject(newObject ValidObject) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.AddObject")
	if newObject == nil {
		return nil, errors.New("newObject can't be nil")
	}
	if reflect.ValueOf(newObject).Kind() != reflect.Ptr {
		return nil, errors.New("check newObject is reference")
	}

	err := newObject.Validate()
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	isExists, err := this.IsExists(newObject)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	if isExists {
		return nil, errors.New("ALREADY_EXISTS")
	}

	err = newObject.CheckBeforeAdd(this)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	return this.Save(newObject)
}

func (this ManagerImpl) UpdateObject(newObject ValidObject) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.UpdateObject")
	if newObject == nil {
		return nil, errors.New("newObject can't be nil")
	}
	if reflect.ValueOf(newObject).Kind() != reflect.Ptr {
		return nil, errors.New("check newObject is reference")
	}

	err := newObject.Validate()
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	isExists, err := this.GetIfExists(newObject)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	if !isExists {
		return nil, errors.New("NOT FOUND UPDATED OBJECT BY KEY")
	}

	err = newObject.CheckBeforeUpdate(newObject, this)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	return this.Save(newObject)
}

func (this ManagerImpl) GetObject(expectedObject ValidObject, keys []string, filterArray...string) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.GetObject %s", keys)
	return this.GetObjectByKeysWithFilter(expectedObject, keys, filterArray...)
}

func (this ManagerImpl) GetObjectByCompositeKey(expectedObject ValidObject, compositeKey string) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.GetObjectByCompositeKey %s", compositeKey)
	bytes, err := this.GetState(compositeKey)
	if err != nil {
		return nil, GlobalLogger.Errorf("GetState by key %s. Error: %s", compositeKey, err)
	}

	if bytes == nil || len(bytes) == 0 {
		GlobalLogger.Error("NOT_FOUND")
		return nil, nil
	}

	err = json.Unmarshal(bytes, expectedObject)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}
	return expectedObject, nil
}

func (this ManagerImpl) GetObjectByKeys(expectedObject ValidObject, keys []string) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.GetObjectByKeys %s", keys)

	validObjects, err := this.GetObjectListByKeys(expectedObject, keys)
	if len(validObjects) != 1 {
		return nil, GlobalLogger.Errorf("Failed GetObjectByKeys found %d expected 1 by keys %s", len(validObjects), keys)
	}

	if err != nil {
		return nil, err
	}

	return validObjects[0], nil
}

func (this ManagerImpl) GetObjectByKeysWithFilter(expectedObject ValidObject, keys []string, filterArray...string) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.GetObjectByKeysWithFilter %s", keys)

	validObjects, err := this.GetObjectListByKeysWithFilter(expectedObject, keys, filterArray...)
	if err != nil {
		return nil, err
	}

	if len(validObjects) > 1 {
		return nil, GlobalLogger.Errorf("Failed GetObjectByKeysWithFilter found %d expected 1 by keys %s", len(validObjects), keys)
	}

	if len(validObjects) == 0 {
		return nil, nil
	}

	return validObjects[0], nil
}

func (this ManagerImpl) PrepareInvokeChaincodeParam(channelName string, chaincodeName string, functionName string, arg interface{}) ([][]byte, error) {
	preference, err := actions.FindSecurityPreferenceByFunctionName(chaincodeName, functionName)
	if err != nil {
		return nil, err
	}

	signedRequest := requests.CCFullSignedRequest{
		UserSignature:   requests.UserSignature{},
		SystemSignature: requests.UserSignature{},
		Parameters: requests.CCParameters{
			UserParams: requests.CCParameter{},
			SystemParams: requests.CCParameter{
				UserID:   "NOT_APPLICABLE",
				ActionId: preference.GetActionID(channelName),
				Param:    arg,
			},
		},
	}

	param, err := json.Marshal(signedRequest)
	if err != nil {
		return nil, err
	}

	return [][]byte{[]byte(preference.FunctionName), param}, nil
}

func (this ManagerImpl) SignedInvokeChaincode(channelName string, chaincodeName string, functionName string, arg interface{}) pb.Response {
	GlobalLogger.Infof("managerImpl.SignedInvokeChaincode")

	bytes, err := this.PrepareInvokeChaincodeParam(channelName, chaincodeName, functionName, arg)
	if err != nil {
		return shim.Error(err.Error())
	}

	return this.InvokeChaincode(chaincodeName, bytes, channelName)
}

func (this ManagerImpl) InvokeChaincode_GetUser(UserID string) pb.Response {
	GlobalLogger.Infof("managerImpl.InvokeChaincode_GetUser")
	return this.SignedInvokeChaincode(invoke.UserChannel, invoke.UserCC, invoke.GetUserFunction, requests.Id{Id: UserID})
}

func (this ManagerImpl) InvokeChaincode_VerifyUserSignature(arg interface{}) pb.Response {
	GlobalLogger.Infof("managerImpl.InvokeChaincode_VerifyUserSignature")
	return this.SignedInvokeChaincode(invoke.UserChannel, invoke.UserCC, invoke.VerifyUserSignatureFunction, arg)
}