package transactionCC

import (
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
	"fmt"
	"errors"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"encoding/json"
	"reflect"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/remote/remote_systempreference"
	"strconv"
)

// TransactionRecord transaction struct which denotes transaction
// fields which are necessary for processing of
// transactions and will be used to present transaction in the ledger
type TransactionRecord struct {
	Key                        string                     `json:"key"`
	OperationID                string                     `json:"operation_id,omitempty"`
	ProcessType                ProcessType                `json:"process_type,omitempty"`
	OperationType              OperationType              `json:"operation_type,omitempty"`
	UserAccountInfo            *accountCC.UserAccountInfo `json:"user_account_info,omitempty"`
	DestinationUserAccountInfo *accountCC.UserAccountInfo `json:"destination_user_account_info,omitempty"`
	Amount                     nums.PlainCurrencyNumber   `json:"amount,omitempty"`
	HashCode                   string                     `json:"hash_code,omitempty"`
	DependencyKey              string                     `json:"dependency_key,omitempty"` // transaction dependency in a form of "txnType.currencyType.accountId"
	Timestamp                  Timestamp                  `json:"timestamp"`
	InOutType                  InputOutputType            `json:"input_output_type"`
}

func (this TransactionRecord) GetObjectType() string  {
	return "TransactionRecord";
}

func (this TransactionRecord) CheckBeforeAdd(manager chaincode.IManager) error {
	// TODO добавить проверку в регуляторном канале передав this
	// проверка статических переменных
	// общие данные в общем канале
	// время добавления
	// кто добавления
	// время обновления
	// кто обновил

	// проверка типа процесса и доступных криптовалют для этого процесса
	processTypeCryptocurrencyRemoteManager := remote_systempreference.GetSystempreferenceRemoteManager(manager, invoke.Processtypecryptocurrencychannel, invoke.ProcesstypecryptocurrencyCC)
	processPreference, err := processTypeCryptocurrencyRemoteManager.GetPreference(&requests.Id{Id: this.ProcessType.String()})
	if err != nil {
		return err
	}
	if processPreference == nil {
		return chaincode.GlobalLogger.Errorf("Process type %s not found", this.ProcessType.String())
	}
	processCurrencyType := processPreference.Fields[this.Amount.CurrencyType]
	if processCurrencyType != "true" {
		return chaincode.GlobalLogger.Errorf("currency type %s is not avalible for process type %s", this.Amount.CurrencyType, this.ProcessType.String())
	}

	err = this.Amount.ValidateAmount()
	if err != nil {
		return err
	}

	isAmountEmpty, err := this.Amount.IsAmountEmpty()
	if err != nil {
		return err
	}
	if isAmountEmpty {
		return chaincode.GlobalLogger.Error("amount can't be 0")
	}

	// проверка типа криптовалюты
	systemPreferenceRemoteManager := remote_systempreference.GetSystempreferenceRemoteManager(manager, invoke.Cryptocurrencyconstraintchannel, invoke.CryptocurrencyconstraintCC)
	cryptocurrencyConstraintPreference, err := systemPreferenceRemoteManager.GetPreference(&requests.Id{Id: this.Amount.CurrencyType})
	if err != nil {
		return err
	}
	if cryptocurrencyConstraintPreference == nil {
		return chaincode.GlobalLogger.Errorf("Constraints for CurrencyType %s not found", this.Amount.CurrencyType)
	}

	scaleUint64, err := strconv.ParseUint(cryptocurrencyConstraintPreference.Fields["scale"], 10, 32)
	if err != nil {
		return chaincode.GlobalLogger.Error(err.Error())
	}
	scale := uint(scaleUint64)

	newAsset, err := nums.NewAsset(cryptocurrencyConstraintPreference.Key, scale, cryptocurrencyConstraintPreference.Fields["min_value"], cryptocurrencyConstraintPreference.Fields["max_value"])
	number, err := this.Amount.Number()
	if err != nil {
		return chaincode.GlobalLogger.Error(err.Error())
	}

	number.Asset = newAsset
	_, err = number.Validate()
	if err != nil {
		return chaincode.GlobalLogger.Error(err.Error())
	}
	return nil
}

func (this TransactionRecord) CheckBeforeUpdate(updatedObject chaincode.ValidObject, manager chaincode.IManager) error {
	return errors.New("METHOD_DOES_NOT_IMPLEMENT")
}

func (this TransactionRecord) GetKey(manager chaincode.IManager) []string {
	return []string{this.UserAccountInfo.UserID, this.UserAccountInfo.WalletID, this.UserAccountInfo.AccountID, this.Amount.CurrencyType, this.InOutType.String(), this.OperationID, this.OperationType.String(), manager.GetTxID()}
}

func (this TransactionRecord) New() chaincode.ValidObject {
	return &TransactionRecord{}
}

func (this TransactionRecord) Validate() error {

	if len(this.DependencyKey) == 0 {
		return errors.New("DependencyKey is empty")
	}

	if this.UserAccountInfo == nil {
		return errors.New("UserAccountInfo is empty")
	}

	if len(this.UserAccountInfo.UserID) == 0 {
		return errors.New("UserID is empty")
	}

	if len(this.Amount.CurrencyType) == 0 {
		return errors.New("CurrencyType is empty")
	}

	if len(this.OperationID) == 0 {
		return errors.New("OperationID is empty")
	}

	if len(this.InOutType) == 0 {
		return errors.New("InOutType is empty")
	}

	if this.UserAccountInfo == nil {
		errMsg := fmt.Sprintf("UserAccountInfo can't be nil")
		return errors.New(errMsg)
	}

	err := this.UserAccountInfo.Validate()
	if err != nil {
		return err
	}

	// для входяших транзакций к пользователю на блокировку нужно делать проверку для поля DestinationUserAccountInfo
	// для остальных операций этого поля НЕТ
	if this.OperationType == LockAmount && this.InOutType == INPUT {

		if this.DestinationUserAccountInfo == nil {
			return errors.New("DestinationUserAccountInfo can't be nil")
		}

		err = this.DestinationUserAccountInfo.Validate()
		if err != nil {
			return errors.New("DestinationUserAccountInfo. " + err.Error())
		}
	}

	err = this.Amount.Validate()
	if err != nil {
		return err
	}

	return nil
}

func Unmarshal(data []byte) (*TransactionRecord, error) {
	if data == nil {
		return nil, chaincode.GlobalLogger.Error("can't Unmarshal nil data")
	}

	var obj TransactionRecord
	err := json.Unmarshal(data, &obj)
	return &obj, err;
}

func Marshal(data *TransactionRecord) ([]byte, error) {
	return json.Marshal(data);
}

func Cast(validObject chaincode.ValidObject) (*TransactionRecord, error) {
	if validObject == nil {
		return nil, nil
	}

	if reflect.ValueOf(validObject).Kind() != reflect.Ptr {
		return nil, errors.New("check validObjectc is not reference")
	}

	casted, ok := validObject.(*TransactionRecord)
	if !ok {
		return nil, errors.New("Problem to get object")
	}

	return casted, nil;
}
